##
## BrubeckService handlers 
##
import logging
from brubeck.request_handling import MessageHandler
from brubeckservice.coro import (
    coro_conn,
    coro_count,
    coro_spawn,
)
# supported methods for a service (MAYBE, NOT ALL IMPLEMENTED)
_SERVICE_METHODS = ['get', 'post', 'put', 'delete', 
                   'options', 'connect', 'response', 'request']

class ServiceMessageHandler(MessageHandler):
    """Class is the simplest implementation of a message handlers. 
    Intended to be used for Service inter communication.
    """
    def __init__(self, application, message, *args, **kwargs):
        self.headers = {}
        super(ServiceMessageHandler, self).__init__(application, message, *args, **kwargs)
        
    def render(self, status_code=None, status_msg=None, headers = None, **kwargs):
        if status_code is not None:
            self.set_status(status_code, status_msg)

        if headers is not None:
            self.headers = headers

        body = self._payload
        
        logging.info('%s %s %s (%s:%s) for (%s:%s)' % (self.status_code, self.message.method,
                                        self.message.path,
                                        self.message.sender,
                                        self.message.conn_id,
                                        self.message.origin_out_addr,
                                        self.message.origin_conn_id,
                                        ))

        return body

    def __call__(self):
        """This is similar to the base call in MessageHandler without the 
        assumption we are an HTTP request.

        It requires a method attribute to indicate which function on the
        handler should be called. If that function is not supported, call the
        handlers unsupported function.

        In the event that an error has already occurred, _finished will be
        set to true before this function call indicating we should render
        the handler and nothing else.

        In all cases, generating a response for the service request is attempted.
        """
        try:
            self.prepare()
            if not self._finished:
                mef = self.message.method.lower()  # M-E-T-H-O-D man!

                # Find function mapped to method on self
                if (mef in _SERVICE_METHODS):
                    fun = getattr(self, mef, self.unsupported)
                else:
                    fun = self.unsupported

                # Call the function we settled on
                try:
                    if not hasattr(self, '_url_args') or self._url_args is None:
                        self._url_args = []

                    if isinstance(self._url_args, dict):
                        ### if the value was optional and not included, filter it
                        ### out so the functions default takes priority
                        kwargs = dict((k, v)
                                      for k, v in self._url_args.items() if v)
                        rendered = fun(**kwargs)
                    else:
                        rendered = fun(*self._url_args)

                    if rendered is None:
                        logging.debug('Handler had no return value: %s' % fun)
                        return ''
                except Exception, e:
                    logging.error(e, exc_info=True)
                    rendered = self.error(e)

                self._finished = True
                return rendered
            else:
                return self.render()
        finally:
            self.on_finish()

def ServicesHandler(MessageHandler):
    def get(self):
        data  = [{
            "name": "services",
            "coro_count": coro_count(self.application),
            "coro_count_service_in": coro_count(self.application, 'service_in'),
            "coro_count_service_out": coro_count(self.application, 'service_out'),
            "coro_count_heartbeat": coro_count(self.application, 'heartbeat'),
            "coro_count_heartbeart_listener": coro_count(self.application, 'heartbeat_listener'),
            "coro_count_registration": coro_count(self.application, 'registration'),
            "coro_count_registration_listener": coro_count(self.application, 'registration_listener'),
        }]
        return self.render_template('services', **data)


"""This is an abstraction for a directory of resources scoped at the application class level.
Since a Brubeck class is generally run in it's own instance, in effect behaving like a singleton, 
it essentially manages resource for a single Brubeck instance. 
An example of a resource could be a Brubeck service (brubeckservice.models.ServiceInfo)"""
import logging
import time
from dictshield.document import Document
from dictshield.fields import (StringField,
                               DictField,
)

# Used to hold our resources for the life of the application
# This will be a reference to a dict stored in a property of the application 
# class that will be added dynamically
_resources = None
_resource_listeners = None

def assure_resource(application=None):
    """assure that our application, if it exists, has a _resources property"""
    logging.debug("Checking for %s _resources " % application)
    global _resources
    if application is None:
        _resources = {}
        _resource_listeners = {}
        logging.debug("_resources initialized and stored in request scope only " % _resources)
        
    elif not hasattr(application, '_resources'):
        setattr(application, '_resources', {})
        setattr(application, '_resource_listeners', {})
        logging.debug("Created _resource attr")
        _resources = application._resources
        _resources = application._resources
    else:
        logging.debug("application._resources (%s) already initialized" % application._resources)
        _resource_listeners = application._resources
        _resources = application._resources

def is_resource_registered(key):
    """ Check if a resource is registered.""" 
    if not _resources is None and key in _resources:
        return True
    else:
        return False

def register_resource(resource, key=None):
    """ Store (register) a resource and call our on_register hook.
        A resource can be anything. Tuples prefered, objects OK.
        key can be None only if we are using a Resource object.
    """

    if isinstance(resource, Resource):
        key = resource.key()
            
    if key not in _resources:
        # add us to the list
        _resources[key] = resource
        # call our hook
        if isinstance(resource, Resource):
            resource._on_register()
        #logging.debug("register_resource success key: %s" % key)
        return True
    else:
        #logging.debug("register_resource ignored: already registered %s" % key)
        return False
        

def unregister_resource(key, sender_id=None):
    """ Call our on_unregister hook and delete from list.
    """ 
    if key not in _resources:
        #logging.debug("unregister_resource ignored: %s not registered" % key)
        return False
    else:
        resource = _resources[key]
        #remove us from the list
        del _resources[key]
        # call our hook
        if isinstance(resource, Resource):
            resource._on_unregister()
        #logging.debug("unregister_resource instance success: %s" % key)   
        return True

def get_resource(key):
    """ Check if key is a resource that is registered
    and return resource if found, None if not found.
    """ 
    if not _resources is None and key in _resources:
        #logging.debug("get_resource found resource %s" % key)
        return _resources[key]
    else:
        # logging.debug("get_resource didn't find resource %s" % key)
        return None

def get_resource_keys():
    """ get our list of resource keys.
    """ 
    if _resources is None:
        return []
    return _resources.keys()
                    
def create_resource_key(name, resource_type=None):
    """used as the key when a resource is stored"""
    return "%s:%s" % ((resource_type if resource_type is not None else ''),
            (name if name is not None else ''))


###################################################
## Moved away from classes for now, not really used
###################################################
class Resource(Document):
    """A resource is an application level object to store arbitrary data in.
        A resource may implement the following methods few methods:
        on_registered(application)
        `on_registered` is called once when a `resource` is registered.
        on_unregistered(application)
        `on_unregistered` is called once when a `resource` is unregistered.
        
        set(resource)
        sets a resource for the instance
        get()
        retrun the last `resource` that was `set(resource)`
        
        
        init expects `resource` as a keyword parameter
        """
    # The name of the resource. This is what it is looked up by.
    name = StringField(required=True)
    # a resource_type such as data_connection, queryset, service, anything really
    resource_type = StringField(required=True)

    sender_id = StringField(required=True)
    
    def __init__(self, *args, **kwargs):
        self.resource = None # this is where we store the user defined resource
        self.start_timestamp = int(time.time() * 1000)
        self.end_timestamp = self.start_timestamp
        super(Resource, self).__init__(*args, **kwargs)
    
    
    def _on_register(self):
        self.on_unregister()
    
    def _on_unregister(self):
        self.on_unregister()
    
    def set(self, resource):
        self.resource = resource;
    
    def get(self):
        return self.resource;
    
    def key(self):
        """used as the key when a resource is stored"""
        return create_resource_key(self.name, self.resource_type)
        
    ###########
    ## over-ride these in your app if you need
    ## to do something durinsg registration/unregistration
    ###########
    def on_register(self):
        pass
    
    def on_unregister(self,):
        pass



############################################
## Service registration (Resource) helpers
############################################
import logging
import time
import ujson as json

from resource import (
    assure_resource,
    is_resource_registered,
    register_resource,
    unregister_resource,
    get_resource,
    create_resource_key,
    get_resource_keys,
    Resource,
)
from brubeck.request_handling import (
    CORO_LIBRARY,
)
from tnetstrings import (
    t_parse,
    t,
)
from models import (
    ServiceRequest,
    ServiceResponse,
)

from coro import (
    coro_get_event,
    coro_send_event,
    coro_sleep,
    CORO_LIBRARY,
    coro_conn,
    coro_count,
    coro_spawn,
)
  
_SERVICE_RESOURCE_TYPE = 'SERVICE'
_DEFAULT_SERVICE_REQUEST_METHOD = 'request'
_DEFAULT_SERVICE_RESPONSE_METHOD = 'response'

REQUEST_TYPE_SYNC = 0 # Request is sent and sender waits for response
REQUEST_TYPE_ASYNC = 1 # Request is sent and sender continues on
REQUEST_TYPE_FORWARD = 2 # Request is sent and sender cedes any responsabilities

HANDLE_RESPONSE = 1
DO_NOT_HANDLE_RESPONSE = 0

def _service_is_registered(application, service_id):
    """ Check if a service is registered"""
    key = create_resource_key(service_id, _SERVICE_RESOURCE_TYPE)
    return is_resource_registered(key)

def _register_service(application, 
    service_passphrase, service_id, service_addr, 
    service_resp_addr, service_conn):
    """ Create and store a connection and it's listener and waiting_clients queue.
    """
    assure_resource(application)      
    sender_id = service_conn.sender_id
    key = create_resource_key(service_id, _SERVICE_RESOURCE_TYPE)  
    if not is_resource_registered(key):                
        instances = {}
        instance = {            
            'sender_id': sender_id,
            'service_conn': service_conn, 
            'waiting_clients': {}, 
        }
        instances[sender_id] = instance
        
        # add us to the list
        resource = {
            'resource_type': _SERVICE_RESOURCE_TYPE,
            'current_sender_id': sender_id,
            'service_id': service_id ,
            'instances': instances, 
            'timestamp': int(time.time()) 
        }

        # create and start our listener
        logging.debug("_register_service starting service response listener (service_id, service_addr, sender_id): (%s,%s,%s)" % (service_id, service_addr, sender_id))        
        if register_resource(resource, key):
            # Start our listener to receive service responses
            coro_spawn(
                service_response_listener,
                application,
                'registration_listener',
                service_id, 
                service_addr, 
                service_resp_addr, 
                service_conn, 
                service_passphrase
            )
            #logging.debug("_register_service success: %s" % key)
            # give above process a chance to start
            coro_sleep(0)
        else:
            logging.debug("_register_service failure: %s" % key)

    else:
        # we may just want to add our instance
        resource = get_resource(key)
        instances = resource['instances']
        sender_id = service_conn.sender_id
        
        if instances is None:
            instances = {}

        if not sender_id in instances:
            #logging.debug("adding %s instance: %s " % (service_id, sender_id))
            instance = {            
                'sender_id': sender_id ,
                'service_conn': service_conn, 
                'waiting_clients': {}, 
            }
            instances[sender_id] = instance 

        resource['current_sender_id'] = service_conn.sender_id
            
        logging.debug("register_service service_conn reconnect(%s:%s:%s)" % (
            service_conn.in_addr, service_conn.out_addr, service_conn.sender_id))

        #logging.debug("register_service  service_conn.sender_id: %s" %
        # (service_conn.sender_id))
        #logging.debug("register_service  application.msg_conn.sender_id: %s" %
        # (application.msg_conn.sender_id))

        # Start our listener to receive service responses
        coro_spawn(
            service_response_listener,
            application,
            'registration_listener',
            service_id, 
            service_addr, 
            service_resp_addr, 
            service_conn, 
            service_passphrase
        )

        service_conn.reconnect(service_conn.sender_id)
        logging.debug("register_service ignored: %s already registered" % service_id)
    return True

def _unregister_service(application, service_id, sender_id):
    """ unregister a service.
    """
    if not _service_is_registered(application, service_id):
        logging.debug("unregister_resource ignored: (%s,%s) not registered" % (service_id, sender_id))
        return False
    else:
        
        service_info = _get_service_info(application, service_id, sender_id)

        if not service_info is None:
            sender_id = service_info['sender_id']
            service_conn = service_info['service_conn']
            waiting_clients = service_info['waiting_clients']
        else:
            logging.debug("unregister_resource ERROR: (%s,%s) no instances." % (service_id, sender_id))
        conn = None
        # should pause a bit here waiting for final responses
        # would need to
        #for sock in waiting_clients:
        #    sock.close()
                        
        key = create_resource_key(service_id, _SERVICE_RESOURCE_TYPE)    

        resource = get_resource(key)
        instances = resource['instances']
        #service_conn = resource['service_conn']
        
        waiting_clients = service_info['waiting_clients']
        conn = service_info['service_conn']
        if not conn is None:
            #logging.debug("conn closing incoming")
            close_in(conn)
            #logging.debug("conn waiting for current clients to finish: %s" % len(waiting_clients))
            stime = int(time.time())
            timeout = 5
            while  (int(time.time()) - timeout) < stime and len(waiting_clients) > 0:
                coro_sleep(1)
            #logging.debug("conn closing outgoing")
            close_out(conn)
        logging.debug("removing instance: %s" % sender_id)                    
        del instances[sender_id]


        if len(instances) == 0:
            # really unregister us
            unregister_resource(key, sender_id)
            logging.debug("unregister_resource success: %s" % key)   
            #service_conn.reconnect(sender_id)
            if isinstance(service_info, Resource):
                service_info._on_unregistered()
        else:
            #logging.debug("unregister_resource service_conn: %s" % 
            # ( id(service_conn) ))
            #logging.debug("unregister_resource conn: %s" % (id(conn)))
            
            logging.debug("unregister_resource conn == service_conn: %s" % (conn == service_conn))                    
            if conn == service_conn:
                #logging.debug("unregister_resource reassinging service_conn: %s" % (key))                    
                resource['service_conn']=instances.itervalues().next()
            logging.debug("unregister_service partial success: %s:%s" % (service_id, len(instances)))
            
        return True

## message listener
def service_response_listener(application, service_id, service_addr, service_resp_addr, service_conn, service_passphrase):
    """Function runs in a coroutine, one listener for each server per handler.
    Once running, it stays running until the brubeck instance is killed."""
    try:
        logging.debug("service_response_listener: (service_id, service_addr, service_resp_addr,service_passphrase): (%s:%s:%s:%s)" % (
            service_id, service_addr, service_resp_addr, service_passphrase))
        loop = True
        key = create_resource_key(service_id, _SERVICE_RESOURCE_TYPE)  
        while is_resource_registered(key) and loop == True:
            logging.debug("service_response_listener waiting")
            raw_response = service_conn.recv()
            logging.debug("service_response_listener recv(): %s" % raw_response)
            # just send raw message to connection client
            if raw_response is None:
                loop = False
            else:
                sender, conn_id = raw_response.split(' ', 1)
                conn_id = t_parse(conn_id)[0]
                sender_id = sender
                _notify_waiting_service_client(application, service_id, conn_id, raw_response, sender_id)
                # Call our handler
                (response, handler_response) = service_conn.process_message(
                    application,
                    raw_response,
                    service_addr,
                    service_passphrase,
                )
        service_conn.close()

        logging.debug("service_response_listener ending: (service_id, service_addr, service_resp_addr,service_passphrase): (%s:%s:%s:%s)" % (
            service_id, service_addr, service_resp_addr, service_passphrase))
        
    except Exception, e:
        logging.error(e, exc_info=True)

# functions for parsing messages
def parse_service_request(msg, passphrase):
    """Function for constructing a Request instance out of a
    message read straight off a zmq socket from a ServiceClientConnection.
    """
    if msg is None:
        return None
        
    #logging.debug("parse_service_request: %s" % msg)
    fields = (sender, conn_id, request_timestamp, msg_passphrase, 
    origin_sender_id, origin_conn_id, origin_out_addr, handle_response, 
    path, method, rest) = msg.strip().split(' ', 10)
    # first field is not tnetstring, no need to do anything
    # last is group of tnetstrings, will handle after
    i=1
    for field in fields[1:-1]:
        fields[i] = t_parse(field)[0]
        i+=1
    # our minimal "security" 
    if fields[3] != passphrase:
        raise Exception('Unknown service identity! (%s != %s)' % (str(fields[3]),str(passphrase)))
    # handle the "body" of the message or last parsed field
    arguments, rest = t_parse(rest)
    headers, rest = t_parse(rest)
    body = t_parse(rest)[0]
    arguments = json.loads(arguments) if len(arguments) > 0 else {}
    headers = json.loads(headers) if len(headers) > 0 else {}
    body = json.loads(body) if len(body) > 0 else {}
    # create our sevice request object
    r = ServiceRequest(**{
            "sender": fields[0],
            "conn_id": fields[1],
            "request_timestamp": fields[2],
            "origin_sender_id": fields[4],
            "origin_conn_id": fields[5],
            "origin_out_addr": fields[6],
            "handle_response": fields[7],
            "path": fields[8],
            "method": fields[9],
            "arguments": arguments,
            "headers": headers,
            "body": body,
    })
    return r

#############################################
## Functions for sending the service request
#############################################
def _send_service_request(application, service_id, service_req):
    """send our message, used internally only.
    returns a tuple with the first element the sender_id and the second the response"""
    service_info = _get_service_info(application, service_id)
    sender_id = None
    if not service_info is None:
        #for key in service_info.keys():
        #    logging.debug("service_info key: %s" % key)
        if 'current_sender_id' in service_info:
            sender_id = service_info['current_sender_id']
        elif 'sender_id' in service_info:
            sender_id = service_info['sender_id']
    #logging.debug("sending service request")
    service_conn = _get_service_conn(application, service_id, sender_id)
    #logging.debug("service_conn: %s"  % service_conn)
    return (sender_id, service_conn.send(service_req))


def _get_service_info(application, service_id, sender_id=None, timeout=10):
    """every internal function needs to use this to get a service_info.
    will wait for a service_info if there is none registered."""
    logging.debug("_get_service_info application.msg_conn.sender_id, sender_id: %s,%s" % 
        (application.msg_conn.sender_id, sender_id))
    service_info = None
    key = create_resource_key(service_id, _SERVICE_RESOURCE_TYPE)
    if _service_is_registered(application, service_id):
        service_info = get_resource(key)
        if sender_id is None:
            return service_info
    else:
        timestart = int(time.time())
        time_timeout = (timestart + timeout)
        now = int(time.time())
        service_info = None
        #logging.debug('timestart + timeout: %s + %s = %s' % (timestart, timeout, time_timeout))
        #logging.debug('now: %s' % (now))
        while service_info is None and now < time_timeout:
            service_info = get_resource(key)
            if(service_info is None):
                coro_sleep(1)
                logging.debug('waiting for service_info (%s): %s' % (key, now))
                now = int(time.time())
        if timeout >0:
            if service_info is None:
                logging.debug("_get_service_info timed out (%s)" % timeout)
            else:
                logging.debug("_get_service_info success waited %s" % (now-timestart))
        elif service_info is None:
            logging.debug("_get_service_info service_info not found")

    if not service_info is None:
        instances = service_info['instances']

        if sender_id is None or not sender_id in instances:
            sender_id = service_info['current_sender_id']
            logging.debug('using current sender_id: %s' % sender_id)
                
        if sender_id in instances:
            instance = instances[sender_id]
            return instance

    return service_info
        #raise Exception("%s service not registered" % service_id)

def _get_service_conn(application, service_id, sender_id=None, timeout=10):
    """get the ServiceClientConnection for a service.
    uses _get_service_info, so it will wait for a resource if there is none initialy.
    """
    service_info = _get_service_info(application, service_id, sender_id)
    if service_info is None:
        # logging.debug("_get_service_conn service_info not found")
        return None
    else:
        # logging.debug("_get_service_conn service_conn: %s" % service_info['service_conn'])
        return service_info['service_conn'] 

def get_service_keys():
    """ get our list of resources.
    """ 
    _resource_keys = get_resource_keys()
    _service_keys = []
    for key in _resource_keys:
        key_parts = key.split(":")
        if key_parts[0]==_SERVICE_RESOURCE_TYPE:
            _service_keys.append(key)
    return _service_keys


# functions for creating messages
def create_service_response(service_request, handler, method=_DEFAULT_SERVICE_REQUEST_METHOD, arguments={}, msg={}, headers={}):
    """Function for creating a ServiceResponse object to send."""
    if not isinstance(headers, dict):
        headers = json.loads(headers)
    if not isinstance(msg, dict):
        msg = json.loads(msg)

    service_response = ServiceResponse(**{
        "sender": service_request.sender,
        "conn_id": service_request.conn_id,
        "origin_sender_id": service_request.origin_sender_id,
        "origin_conn_id": service_request.origin_conn_id,
        "origin_out_addr": service_request.origin_out_addr,
        "handle_response": service_request.handle_response,
        "path": service_request.path,
        "method": method,
        "arguments": arguments,
        "headers": headers,
        "body": msg,
        "status_code": handler.status_code,
        "status_msg": handler.status_msg,
    })

    return service_response
def _notify_waiting_service_client(application, service_id, conn_id, raw_results, sender_id):
    """Notify waiting events if they exist."""
    waiting_clients = _get_waiting_service_clients(application, service_id, sender_id)
    logging.debug("waiting_clients: %s" % (waiting_clients))
    conn_id = str(conn_id)
    if not waiting_clients is None and conn_id in waiting_clients:
        logging.debug("conn_id %s found to notify(%s)" % (conn_id,waiting_clients[conn_id]))
        coro_send_event(waiting_clients[conn_id][1], raw_results)
        del waiting_clients[conn_id]
        #logging.debug("conn_id %s sent to: %s" % (conn_id, raw_results))
        coro_sleep(0)
        return True
    else:
        logging.debug("conn_id %s not found to notify." % conn_id)
        return False

def _get_waiting_service_clients(application, service_id, sender_id):
    """get the waiting service clients."""
    service_info = _get_service_info(application, service_id, sender_id)
    if service_info is None:
        return None
    else:     
        return service_info['waiting_clients']

## Connection functions

def close(conn):
    """close all connections"""
    conn.out_sock.close()
    conn.in_sock.close()

def close_out(conn):
    """close our outgoing connection"""
    conn.out_sock.close()

def close_in(conn):
    """close our incoming connection"""
    conn.in_sock.close()


import os
class ServiceError(Exception):
    """Base class for errors in the brubeckservice package."""
    def __init__(self, message, Errors):

        # Call the base class constructor with the parameters it needs
        Exception.__init__(self, message)

        # Now for your custom code...
        self.Errors = Errors


# helper functions for creating/parsing tnetstrings fields
def t(text):
    """create a tnetstring field given the text"""
    return "%d:%s" % (len(str(text)), str(text))

def t_parse(field_text):
    """ parse a tnetstring field, and return any remainder
        field_value - a value in n:data format where n is the data length
            and data is the text to get the first n chars from
        returns the a tuple containing the value and whatever remains
    """
    #logging.debug("t_parse: %s" % field_text)
    field_data = field_text.split(':', 1)
    expected_len = int(field_data[0])
    #logging.debug("expected_len: %s" % expected_len)
    if expected_len > 0:
        field_value = field_data[1]
        value = field_value[0:expected_len]
    else:
        value = ''
    rest = field_value[expected_len:] if len(field_value) > expected_len else ''
    return (value, rest)



import logging
from brubeck.request_handling import (
    CORO_LIBRARY,
)

################################################################
### Attempt to setup gevent wrappers for sleep, events and more
### Always prefere gevent if installed, then try eventlet
################################################################
coro_pool = {}
if CORO_LIBRARY == 'gevent':
    from gevent import sleep
    from gevent.event import AsyncResult
    from gevent import pool

    def coro_sleep(secs):
        sleep(secs)

    def coro_get_event():
        return AsyncResult()

    def coro_send_event(e, value):
        e.set(value)

    def coro_create_pool():
        return pool.Pool()

    def coro_pool_wait_available(pool):
        if pool is None:
            return 0
        return pool.waiting()

    def _coro_spawn(p, function, app, *a, **kw):
        return p.spawn(function, app, *a, **kw)

    CORO_LIBRARY = 'gevent'

elif CORO_LIBRARY == 'eventlet':
    from eventlet import event
    from eventlet import sleep

    def coro_sleep(secs):
        sleep(secs)

    def coro_get_event():
        return AsyncResult()

    def coro_send_event(e, value):
        e.set(value)

    def coro_create_pool():
        return eventlet.GreenPool()

    def coro_pool_wait_available(pool):
        if pool is None:
            return 0
        return pool.wait_available()

    def _coro_spawn(p, function, app, *a, **kw):
        return p.spawn_n(function, app, *a, **kw)

    CORO_LIBRARY = 'eventlet'

# common coro functions using above wrapper functions
def coro_conn(app, tag = None):
    """get a coro pool"""
    if tag is None or tag == 'coro_pool':
        logging.debug("coro_conn returning app pool.")
        return app.pool
    elif tag in coro_pool:
        logging.debug("coro_conn returning existing pool for %s" % tag)
        return coro_pool[tag]
    else:
        coro_pool[tag] = coro_create_pool()
        logging.debug("coro_conn returning new pool for %s" % tag)
        return coro_pool[tag]

def coro_spawn(function, app, tag = None, *a, **kw):
    """spawn a coro"""
    logging.debug("getting pool for %s" % tag)
    pool = coro_conn(app, tag)
    logging.debug("pool for %s: %s" % (tag, pool))
    return _coro_spawn(pool, function, app, *a, **kw)

def coro_count(app, tag):
    """get the count of a coro pool"""
    p = None
    if tag is None or tag == 'coro_pool':
        return app.pool.wait()
    elif tag in coro_pool:
        return coro_pool_wait_available(coro_pool[tag])
    else:
        return 0



#!/usr/bin/env python
 
from setuptools import setup
 
setup(name='brubeck-service',
      version='0.1.2',
      description='Library to allow Brubeck framework to call other instances as services',
      author='Seth Murphy',
      author_email='seth@brooklyncode.com',
      url='http://github.com/sethmurphy/brubeckservices',
      packages=['brubeckservice'])



#!/usr/bin/env python

"""The Brubeck Services Addon."""

version = "0.1.0"
version_info = (0, 1, 0)
__all__ = ['base, connections, coro, handlers, models, resource, service, ServiceError, tnetstrings']



##
## BrubeckService models
##
import time
from dictshield.document import Document
from dictshield.fields import (
    StringField,
    BooleanField,
    URLField,
    EmailField,
    LongField,
    DictField,
    IntField,
)

class ServiceRequest(Document):
    """Class used to construct a Brubeck service request message.
    Both the client and the server use this.
    """
    # set by the send call in the client connection
    sender = StringField(required=True)
    # set by the send call in the client connection
    conn_id = StringField(required=True)
    # set by send call in client connection
    # used to return the message to the originator
    origin_sender_id = StringField(required=True)
    # This is the connection id used by the originator and is needed for Mongrel2
    origin_conn_id  = StringField(required=True)
    # This is the socket address for the reply to the client
    origin_out_addr  = StringField(required=True)
    # What type of request are we? (sync=1, async=2, forward=3)
    # This is set by the call to send the request
    request_type = None
    # flag to see if response tries to call handler (1=yes)
    handle_response = StringField(required=True)
    # used to route the request
    path = StringField(required=True)
    # used to route the request to the proper method of the handler
    method = StringField(required=True)
    # a dict, used to populat an arguments dict for use within the method
    arguments = DictField(required=False)
    # a dict, these get passed to the method called (use body for complex or large objects)
    headers = DictField(required=False)
    # a dict, this can be whatever you need it to be to get the job done.
    body = DictField(required=True)

    #_service_request_type = None
    
    def __init__(self, *args, **kwargs):
        self.request_timestamp = int(time.time() * 1000)
        super(ServiceRequest, self).__init__(*args, **kwargs)

    def get_argument(self, key, default=None):
        """get's an argument by name"""
        if key in self.arguments:
            return self.arguments[key]
        return default


class ServiceResponse(ServiceRequest):
    """Class used to construct a Brubeck service response message.
    """
    status_code = IntField(required=True)
    status_message = StringField()
    def __init__(self, *args, **kwargs):
        self.start_timestamp = int(time.time() * 1000)
        self.end_timestamp = 0
        super(ServiceResponse, self).__init__(*args, **kwargs)

class ServiceConnectionInfo(Document):
    """Contains all the infor mation needed to set up a service
    heartbeats and all (Not the request itself)"""
    # passphrase needed for service registration
    service_registration_passphrase = StringField(required=True)   
    # service id or name
    service_id = StringField(required=True)
    # the address to register the service for
    service_registration_addr = StringField(required=True)

    # passphrase needed for service call
    service_passphrase = StringField(required=True)    
    # the service request address
    service_addr = StringField(required=True)
    # the service response address
    service_response_addr = StringField(required=True)
    
    # the service heartbeat address
    service_heartbeat_addr = StringField(required=True)
    # the client heartbeat address
    client_heartbeat_addr = StringField(required=True)
    
    # timeout before an endpoint dead
    service_heartbeat_timeout = IntField(required=True)
    # how often to send a heartbeat
    service_heartbeat_interval = IntField(required=True)
    
    # single connection id for pair
    sender_id = StringField(required=True)

    is_reregistration = IntField(required=True, default=0)
    



import logging
import os
import sys
import time
import ujson as json
import imp
import string
from uuid import uuid4
from brubeck.request import Request
from brubeck.request_handling import (
    CORO_LIBRARY,
    Brubeck,
)
from brubeckservice.connections import (
    ServiceConnection,
    _service_client_registration_listener,
    DEFAULT_HEARTBEAT_INTERVAL,
    start_service_client_registration_listener,
)
from brubeckservice.models import (
    ServiceRequest,
    ServiceResponse,
)
from brubeckservice.handlers import (
    ServiceMessageHandler,
)
from brubeckservice.tnetstrings import (
    t_parse,
)
from brubeckservice.service import (
    _register_service,
    _unregister_service,
    _DEFAULT_SERVICE_REQUEST_METHOD,
)
from brubeckservice.resource import (
    assure_resource,
)
from brubeckservice.service import (
    _SERVICE_RESOURCE_TYPE,
    _send_service_request,
    _get_service_info,
    _get_service_conn,
    _get_waiting_service_clients,
    REQUEST_TYPE_SYNC,
    REQUEST_TYPE_ASYNC,
    REQUEST_TYPE_FORWARD,
    HANDLE_RESPONSE,
    DO_NOT_HANDLE_RESPONSE,
)

from brubeckservice.coro import (
    coro_get_event,
    coro_send_event,
    coro_sleep,
    CORO_LIBRARY,
    coro_spawn,
)

#################################
## Request and Response stuff 
#################################


## handler client mixins
class ServiceClientMixin(object):
    """Class adds the functionality to any handler to send messages to a ServiceConnection
    This must be used with a handler or something that has the following attributes:
        self.application
    """

    ################################
    ## The public interface methods
    ## This is all your handlers should use
    ################################

    def register_service(self, service_id, service_addr, service_resp_addr, service_passphrase, sender_id):
        """Public wrapper around _register_service"""
        return _register_service(self.application, service_id, service_addr, service_resp_addr, service_passphrase, service_conn, sender_id)
        
    def unregister_service(self, service_id, service_passphrase, sender_id):
        """Public wrapper around _unregister_service"""
        return _unregister_service(self.application, service_id, service_passphrase, sender_id)
        
    def create_service_request(self, path, handle_response=HANDLE_RESPONSE, 
        method=_DEFAULT_SERVICE_REQUEST_METHOD, arguments={}, msg={}, headers={}):
        """ path - string, used to route to proper handler
            method - used to map to the proper method of the handler
            arguments - dict, used within the method call if needed
            These are not used anymore, but I feel they belong. 

            If not to only hold the original request
                headers - dict, contains the accepted method to call on handler
                msg - dict, the body of the message to process
        """
        if not isinstance(headers, dict):
            headers = json.loads(headers)
        if not isinstance(msg, dict):
            msg = json.loads(msg)

        data = {
            # Not sure if this is the socket_id, but it is used to return the message to the originator
            "origin_sender_id": self.message.sender,
            # This is the connection id used by the originator and is needed for Mongrel2
            "origin_conn_id": self.message.conn_id,
            # This is the socket address for the reply to the client
            "origin_out_addr": self.application.msg_conn.out_addr,
            # used to route the request
            "handle_response": str(handle_response),
            "path": path,
            "method": method,
            "arguments": arguments,
            # a dict, right now only METHOD is required and must be one of:
            # ['get', 'post', 'put', 
            # 'delete','options', 'connect', 
            # 'response', 'request']
            # "headers": headers,
            # a dict, this can be whatever you need it to be.
            "body": msg,
        }
        return ServiceRequest(**data)
    
    def send_service_request(self, service_id, service_req):
        """do some work and wait for the results of the response to handler_response the response from the service
        blocking, waits for handled result.
        """
        logging.debug("send_service_request(service_id=%s, service_req=%s)" % (service_id, service_req))
        #sender_id = self.application.msg_conn.sender_id
        service_req._request_type = REQUEST_TYPE_SYNC
        (sender_id, service_req) = _send_service_request(self.application, service_id, service_req)
        conn_id = service_req.conn_id
        raw_response = _wait(self.application, service_id, conn_id, sender_id)
        service_conn = _get_service_conn(self.application, service_id, sender_id)
        
        (response, handler_response) = service_conn.process_message( 
            self.application, raw_response, 
            service_id, service_conn.passphrase,
        )

        return (response, handler_response)

    def send_service_request_nowait(self, service_id, service_req):
        """defer some work, but still handle the response yourself
        non-blocking, returns immediately.
        """
        service_req._request_type = REQUEST_TYPE_ASYNC
        _send_service_request(self.application, service_id, service_req)
        return

    def forward_to_service(self, service_id, service_req):
        """give up any responsability for the request, someone else will respond to the client
        non-blocking, returns immediately.
        """
        service_req._request_type = REQUEST_TYPE_FORWARD
        _send_service_request(self.application, service_id, service_req)


###################################################
## Functions for waiting and notifying of response
###################################################
def _wait(application, service_id, conn_id, sender_id):
    """wait for the application to create an event from the service listener"""
    raw_response = None
    conn_id = str(conn_id)
    e = coro_get_event()
    waiting_events = _get_waiting_service_clients(application, service_id, sender_id)
    waiting_events[conn_id] = (int(time.time()), e)

    logging.debug("event %s on %s waiting " % (conn_id, sender_id))
    raw_response = e.get()
    logging.debug("event for %s raised, raw_response: %s" % (conn_id, raw_response))


    if raw_response is not None:
        return raw_response
    else:
        logging.debug("_wait for %s NO RESULTS" % conn_id)
        return None

def _update_service_heartbeat(application, service_id, sender_id):
    """Update our hearbeat timestamp in our service_listener for the service"""
    service_info = _get_service_info(application, service_id, sender_id)
    if not service_info is None:
        service_info['timestamp'] = int(time.time())
        return True
    else:
        logging.debug("service_info %s not found to update heartbeat (service_id):" % 
            (service_id))
        return False

## init functions
def service_client_init(application, 
    service_registration_passphrase, service_id, service_registration_addr,
    service_client_heartbeat_addr = None, service_client_heartbeat_interval = DEFAULT_HEARTBEAT_INTERVAL):
    """Starts everything needed for a brubeck app to be a service_client"""
    logging.debug('******** START service_client_init *********')

    logging.debug('service_registration_passphrase: %s' % service_registration_passphrase)
    logging.debug('service_id: %s' % service_id)
    logging.debug('service_registration_addr: %s' % service_registration_addr)
    logging.debug('service_client_heartbeat_addr: %s' % service_client_heartbeat_addr)
    logging.debug('service_client_heartbeat_interval: %s' % service_client_heartbeat_interval)

    logging.debug('******** END service_client_init *********')
    
    assure_resource(application)
    start_service_client_registration_listener(application,
        service_registration_passphrase, service_id, service_registration_addr, 
        service_client_heartbeat_addr, service_client_heartbeat_interval)

def lazyprop(method):
    """ A nifty wrapper to only load preoperties when accessed
    uses the lazyProperty pattern from:
    http://j2labs.tumblr.com/post/17669120847/lazy-properties-a-nifty-decorator
    inspired by  a stack overflow question:
    http://stackoverflow.com/questions/3012421/python-lazy-property-decorator
    This is to replace initializing common variable from cookies, query string, etc ..
    that would be in the prepare() method.
    """
    attr_name = '_' + method.__name__
    @property
    def _lazyprop(self):
        if not hasattr(self, attr_name):
            attr = method(self)
            setattr(self, attr_name, method(self))
            # filter out our javascript nulls
            if getattr(self, attr_name) == 'undefined':
                setattr(self, attr_name, None)
        return getattr(self, attr_name)
    return _lazyprop

class BrooklynCodeBrubeck(Brubeck):
    """our main application, extending Brubeck
    This is not application specific,
    but general stuff that maybe shoud be in Brubeck someday"""

    def __init__(self, *args, **kwargs):
        """ Most of the parameters are dealt with by Brubeck,
            do a little before call to super
        """
        settings_file = kwargs['settings_file'] if 'settings_file' in kwargs else none
        project_dir = kwargs['project_dir'] if 'project_dir' in kwargs  else none


        if project_dir == None:
            raise Exception('missing project_dir from config')
        else:
            self.project_dir = project_dir

        """load our settings"""
        if settings_file != None:
            self.settings = self.get_settings_from_file(settings_file)
        else:
            self.settings = {}

        # we may have overriden default mongrel2_pairs in settings
        #if 'mongrel2_pair' in self.settings:
        #    kwargs['mongrel2_pair'] = self.settings['mongrel2_pair']

        Brubeck.__init__(self, **kwargs)


    def get_settings(self, setting_name, file_name=None):
        """ This is an attempt at providing a possible
        external location for settings to overide any
        settings in the settings file that was passed
        to the application during initialization.
        """

        try:
            if hasattr(self, 'settings'):
                if hasattr(self.settings, setting_name):
                    # we had our settings loaded already
                    return getattr(self.settings, setting_name)

            if file_name == None:
                # create default file name
                file_name = self.project_dir + '/conf/' + setting_name + '.py'

            # try to load our file
            settings = self.get_settings_from_file(file_name)

            if hasattr(settings, setting_name):
                # load us in self.settings
                # so we don't need to read from file  next time
                self.settings.append({setting_name:settings[settings_name]})
                return settings[settings_name]

            raise Exception("Unable to load settings from file %s: %s " % (file_name, setting_name))

        except:
            # raise our error
            raise


    def get_settings_from_file(self, file_name):

        settings = None
        logging.debug ('loading settings from %s' % file_name)

        try:
            settings = imp.load_source('settings', file_name)
        except:
            raise #Exception("unknown error getting file: " + file_name)

        return settings



##
## BrubeckService connections
##
import logging
import time
from uuid import uuid4
import ujson as json
from brubeck.request import to_bytes, to_unicode
from brubeck.connections import (
    load_zmq,
    load_zmq_ctx,
    Mongrel2Connection,
    Connection,
)
from brubeck.request_handling import (
    CORO_LIBRARY,
)
from brubeck.request_handling import (
    JSONMessageHandler,
    WebMessageHandler, 
    Brubeck,
    render,
)
from brubeckservice.tnetstrings import (
    t,
    t_parse,
)
from brubeckservice.service import (
    _register_service,
    _unregister_service,
    parse_service_request,
    _DEFAULT_SERVICE_RESPONSE_METHOD,
    create_service_response,
    _get_service_info,
    close_out,
    close_in,
)
from brubeckservice.models import (
    ServiceResponse, 
    ServiceConnectionInfo
)
    
from brubeckservice.coro import (
    coro_get_event,
    coro_send_event,
    coro_sleep,
    CORO_LIBRARY,
    coro_spawn,
)

_DEFAULT_SERVICE_CLIENT_TIMEOUT = 5 # seconds before service is reregistered    
_DEFAULT_SERVICE_TIMEOUT = 5 # seconds before service is unregistered on client

_ALLOWED_MISSED_HEARTBEATS = 1 # missed heartbeats before listener killed
DEFAULT_HEARTBEAT_INTERVAL = 3 # time in seconds between heartbeats

class ServiceConnection(Mongrel2Connection):
    """Class is specific to handling communication with a ServiceClientConnection.
    """
    
    def __init__(self, svc_addr, svc_resp_addr, passphrase, sender_id = None):
        """sender_id = uuid.uuid4() or anything unique
        pull_addr = pull socket used for incoming messages
        pub_addr = publish socket used for outgoing messages

        The class encapsulates socket type by referring to it's pull socket
        as in_sock and it's publish socket as out_sock.
        """
        zmq = load_zmq()
        ctx = load_zmq_ctx()

        self.sender_id = uuid4().hex if sender_id is None else sender_id
        self.in_addr = svc_addr
        self.out_addr = svc_resp_addr
        self.last_response_sender_id = None

        # the request (in_sock) is received from a DEALER socket (round robin)
        self.in_sock = ctx.socket(zmq.PULL)
        self.in_sock.connect(self.in_addr)
        print("CONNECT service requested PULL socket %s:%s" % (self.in_addr, self.sender_id))

        # the response is sent to original clients incoming ROUTER socket
        self.out_sock = ctx.socket(zmq.ROUTER)
        self.out_sock.connect(self.out_addr)
        print("CONNECT service response ROUTER socket %s:%s" % (self.out_addr, self.sender_id))

        self.service_heartbeat = None
        self.reg_addr = None
        self.reg_sock = None
        
        self.zmq = zmq
        self.passphrase = passphrase
        
        self.service_heartbeat_connection = None

    def set_service_heartbeat_connection(self, application, service_heartbeat_connection):
        """set and start our service_heartbeat_connection """
        self.service_heartbeat_connection = service_heartbeat_connection
        self.service_heartbeat_connection.start(application)
        
    def process_message(self, application, message):
        """Function for coroutine that looks at the message, determines which handler will
        be used to process it, and then begins processing.
        The application is responsible for handling misconfigured routes.
        """

        service_request = parse_service_request(message, application.msg_conn.passphrase)
        if service_request is None:
            return
            
        handler = application.route_message(service_request)
        result = handler()

        msg = {}

        if result is not None and result is not "":
            msg = json.dumps(result)
        service_response = create_service_response(service_request, handler, method=_DEFAULT_SERVICE_RESPONSE_METHOD, arguments={}, msg=msg, headers={})
        
        application.msg_conn.send(service_response)

    def send(self, service_response):
        """uuid = unique ID that both the client and server need to match
           conn_id = connection id from this request needed to wake up handler on response
           origin_uuid = unique ID from the original request
           origin_conn_id = the connection id from the original request
           origin_out_addr = the socket address that expects the final result
           msg = the payload (a JSON object)
           request_type = what behavior am I expecting from the request (async, sync, forward)
           handle_response = call a handler on response processing
           path = the path used to route to the proper response handler
        """

        service_response.end_timestamp = int(time.time() * 1000)
        
        # check who we last sent a response to
        if self.last_response_sender_id is None:
            # first time, set it
            self.last_response_sender_id = service_response.sender
        elif not service_response.sender == self.last_response_sender_id:
            # Our single client restarted, reconnect
            self.out_sock.connect(self.out_addr)
            self.last_response_sender_id = service_response.sender

        header = "%s %s %s %s %s %s %s %s %s %s %s %s" % ( service_response.sender,
            t(service_response.conn_id),
            t(service_response.request_timestamp),
            t(service_response.start_timestamp),
            t(service_response.end_timestamp),
            t(self.passphrase),
            t(service_response.origin_sender_id),
            t(service_response.origin_conn_id),
            t(service_response.origin_out_addr),
            #t(service_response._service_request_type),
            t(service_response.handle_response),
            t(service_response.path),
            t(service_response.method),
        )
                
        status_code = to_bytes(str(json.dumps(service_response.status_code)))
        status_msg = to_bytes(json.dumps(service_response.status_msg))
        arguments = to_bytes(json.dumps(service_response.arguments))
        headers = to_bytes(json.dumps(service_response.headers))
        body = to_bytes(json.dumps(service_response.body))
        msg = '%s %s%s%s%s%s' % (header,
            t(status_code),
            t(status_msg),
            t(arguments),
            t(headers),
            t(body),
        )
        
        logging.debug("ServiceConnection send (%s:%s) : \"%s\"" % (self.out_addr, service_response.sender, msg))

        self.out_sock.send(service_response.sender, self.zmq.SNDMORE)
        self.out_sock.send("", self.zmq.SNDMORE)
        self.out_sock.send(msg, self.zmq.NOBLOCK)
        return

    def recv(self):
        """Receives a message from a ServiceClientConnection.
        """
        try:
            # blocking recv call
            logging.debug("recv waiting...")
            zmq_msg = self.in_sock.recv()
            logging.debug("...recv got")
            # if we are multipart, keep getting our message until we are done
            while self.in_sock.getsockopt(self.zmq.RCVMORE):
                logging.debug("...recv getting more")
                zmq_msg += self.in_sock.recv()
            logging.debug("...recv got all")

            return zmq_msg
        except Exception, e:
            logging.debug("ServiceConnection recv() going down.")

    def register_service(self, application, 
            service_registration_passphrase, service_id, service_registration_addr,  
            service_passphrase, service_addr, service_response_addr,  
            service_heartbeat_addr, service_heartbeat_timeout=_DEFAULT_SERVICE_CLIENT_TIMEOUT, 
            sender_id = None, 
            is_reregistration = 0):
        """just a wrapper around the function of the same name"""

        return register_service(application, 
            service_registration_passphrase, service_id, service_registration_addr, 
            service_passphrase, service_addr, service_response_addr, 
            service_heartbeat_addr, service_client_heartbeat_addr, 
            service_heartbeat_timeout, sender_id,
            is_reregistration);

    def reconnect(self, sender_id=None):
        """a full reconnect of our underlying sockets"""
        self.in_sock.close()
        self.out_sock.close()

        zmq = load_zmq()
        ctx = load_zmq_ctx()

        if sender_id is None:
            sender_id = uuid4().hex
        self.sender_id = sender_id
        
        self.last_response_sender_id = None

        # the request (in_sock) is received from a DEALER socket (round robin)
        self.in_sock = ctx.socket(zmq.PULL)
        self.in_sock.connect(self.in_addr)
        print("RECONNECT service requested PULL socket %s:%s" % (self.in_addr, self.sender_id))

        # the response is sent to original clients incoming DEALER socket
        self.out_sock = ctx.socket(zmq.ROUTER)
        self.out_sock.connect(self.out_addr)
        print("RECONNECT service response ROUTER socket %s:%s" % (self.out_addr, self.sender_id))


class ServiceClientConnection(ServiceConnection):
    """Class is specific to communicating with a ServiceConnection.
    """

    def __init__(self, application, svc_addr, svc_resp_addr, passphrase, sender_id=None, client_heartbeat_connection = None):
        """ passphrase = unique ID that both the client and server need to match
                for security purposed

            svc_addr = address of the Brubeck Service we are connecting to
            This socket is used for both inbound and outbound messages
        """
        self.active = True
        self.passphrase = passphrase
        self.sender_id = uuid4().hex if sender_id is None else sender_id
        self.out_addr = svc_addr
        self.in_addr = svc_resp_addr
        
        zmq = load_zmq()
        ctx = load_zmq_ctx()

        self.out_sock = ctx.socket(zmq.PUSH)
        self.out_sock.bind(self.out_addr)
        logging.debug("BIND service request PUSH socket %s:%s)" % (self.out_addr, self.sender_id))

        self.in_sock = ctx.socket(zmq.DEALER)
        self.in_sock.setsockopt(zmq.IDENTITY, self.sender_id)
        self.in_sock.setsockopt(zmq.LINGER, 0)
        
        self.in_sock.bind(self.in_addr)
        logging.debug("BIND service response DEALER socket %s:%s" % (self.in_addr, self.sender_id))

        self.service_client_heartbeat = None
        self.reg_sock = None
        self.reg_addr = None
        self.zmq = zmq    

        self.client_heartbeat_connection = client_heartbeat_connection
        if not self.client_heartbeat_connection is None:
            self.client_heartbeat_connection.start(application)

    def set_client_heartbeat_connection(self, application, client_heartbeat_connection):
        """set and start our client_heartbeat_connection """
        self.client_heartbeat_connection = client_heartbeat_connection
        self.client_heartbeat_connection.start(application)

    def process_message(
        self,
        application,
        message,
        service_addr,
        service_passphrase
    ):
        """This coroutine looks at the message, determines which handler will
        be used to process it, and then begins processing.
        Since this is a reply, not a request,
        we simply call the handler and are done
        returns a tuple containing 1) the response object created 
            from parsing the message and 2) the handlers return value
        """
        if message is None:
            return (None, None)
            
        logging.debug("service_client_process_message service_passphrase: %s" % service_passphrase)
        service_response = parse_service_response(message, service_passphrase)
    
        logging.debug(
            "service_client_process_message service_response: %s" % service_response
        )
        
        logging.debug("service_client_process_message handle_response: %s" % service_response.handle_response)
        logging.debug("service_client_process_message service_response.path: %s" % service_response.path)
        if service_response.handle_response == "1":
            handler = application.route_message(service_response)
            
            handler.set_status(service_response.status_code,  service_response.status_msg)
            result = handler()
            logging.debug(
                "service_client_process_message service_response: %s" % service_response)
            logging.debug("service_client_process_message result: %s" % result)
            return (service_response, result)
    
        return (service_response, None)

    def send(self, service_req):
        """Send will wait for a response with a listener and is async
        """
        if service_req is None:
            return None
            
        service_req.conn_id = uuid4().hex

        header = "%s %s %s %s %s %s %s %s %s %s" % (self.sender_id, 
            t(service_req.conn_id), 
            t(service_req.request_timestamp),
            t(self.passphrase),
            t(service_req.origin_sender_id),
            t(service_req.origin_conn_id),
            t(service_req.origin_out_addr),
            t(service_req.handle_response),
            t(service_req.path),
            t(service_req.method),
        )
        arguments = to_bytes(json.dumps(service_req.arguments))
        headers = to_bytes(json.dumps(service_req.headers))
        body = to_bytes(json.dumps(service_req.body))

        msg = '%s %s%s%s' % (header, t(arguments),t(headers), t(body))
        logging.debug(
            "ServiceClientConnection send (%s:%s:%s): %s" % (self.out_addr, self.sender_id, service_req.conn_id, msg)
        )
        self.out_sock.send(to_bytes(msg))

        return service_req

    def reconnect(self, sender_id = None):
        """a full reconnect of our underlying sockets"""
        zmq = load_zmq()
        ctx = load_zmq_ctx()

        self.in_sock.close()
        self.out_sock.close()

        if sender_id is None:
            sender_id = uuid4().hex
        self.sender_id = sender_id

        self.out_sock = ctx.socket(zmq.PUSH)
        self.out_sock.bind(self.out_addr)
        logging.debug("REBIND service request PUSH socket %s:%s" % (self.out_addr, self.sender_id))

        self.in_sock = ctx.socket(zmq.DEALER)
        self.in_sock.setsockopt(zmq.IDENTITY, self.sender_id)
        self.in_sock.bind(self.in_addr)
        logging.debug("REBIND service response DEALER socket %s:%s" % (self.in_addr, self.sender_id))


## message listeners
def register_service(application, 
    service_registration_passphrase, service_id, service_registration_addr,
    service_passphrase, service_addr, service_response_addr,  
    service_heartbeat_addr = None, service_heartbeat_timeout=_DEFAULT_SERVICE_CLIENT_TIMEOUT, 
    sender_id = None, is_reregistration = 0,
    service_heartbeat_interval = DEFAULT_HEARTBEAT_INTERVAL,
    ):
    """Register a service with a client. Called once on service startup to 
    register service with the remote Brubeck service client.
    """

    logging.debug('******** START _service_registration *********')
    logging.debug('service_registration_passphrase: %s' % service_registration_passphrase)
    logging.debug('service_id: %s' % service_id)
    logging.debug('service_registration_addr: %s' % service_registration_addr)
    logging.debug('service_passphrase: %s' % service_passphrase)
    logging.debug('service_addr: %s' % service_addr)
    logging.debug('service_response_addr: %s' % service_response_addr)
    logging.debug('service_heartbeat_addr: %s' % service_heartbeat_addr)
    logging.debug('service_heartbeat_timeout: %s' % service_heartbeat_timeout)
    logging.debug('sender_id: %s' % sender_id)
    logging.debug('is_reregistration: %s' % is_reregistration)
    logging.debug('service_heartbeat_interval: %s' % service_heartbeat_interval)
    logging.debug('******** END _service_registration *********')
    
    service_connection_info = {
            'service_registration_passphrase': service_registration_passphrase,
            'service_id': service_id,
            'service_registration_addr': service_registration_addr,
            'service_passphrase': service_passphrase,    
            'service_addr': service_addr,
            'service_response_addr': service_response_addr,
            'service_heartbeat_addr': service_heartbeat_addr,
            'service_heartbeat_timeout': service_heartbeat_timeout,
            'service_heartbeat_interval': service_heartbeat_interval,
            'sender_id': sender_id,
            'is_reregistration': 1 if is_reregistration==1 else 0,
    }

    logging.debug("register_service: (service_registration_addr, service_registration_passphrase,service_heartbeat_timeout,sender_id): (%s,%s, %s, %s)" % 
        (service_registration_addr, service_registration_passphrase,service_heartbeat_timeout,sender_id))
    # Just start our zmq socket here, no reason for abstraction
    zmq = load_zmq()
    ctx = load_zmq_ctx()

    if sender_id is None:
        sender_id = application.msg_conn.sender_id
    # the request (in_sock) is received from a DEALER socket (round robin)
    service_registration_sock = ctx.socket(zmq.REQ)
    service_registration_sock.connect(service_registration_addr)
    
    print("CONNECT service registration REQ socket %s" % (service_registration_addr))
    msg = "%s %s %s %s %s %s %s" % (sender_id, t(service_registration_passphrase), 
        t(service_id), t(service_addr), t(service_response_addr), t(service_passphrase), 
        t(service_heartbeat_addr))
    logging.debug("service_registration sending request");
    service_registration_sock.send(to_bytes(msg))
    logging.debug("service_registration waiting for response");
    raw_registration_response = service_registration_sock.recv()    
    service_registration_sock.close()
    logging.debug("_service_registration recv(): %s" % raw_registration_response)

    fields = (sndr_id, svc_registration_passphrase, svc_client_heartbeat_addr) = (
        raw_registration_response.split(' ', 2))
    i=1
    service_client_heartbeat_interval = DEFAULT_HEARTBEAT_INTERVAL
    for field in fields[1:]:
        fields[i] = t_parse(field)[0]
        i+=1   
    logging.debug("_service_reg application.msg_conn.reconnect(%s)" % sndr_id)    
    application.msg_conn.reconnect(sndr_id)
    
    # our minimal "security"    
    if fields[1] != service_registration_passphrase:
        # just log the breach
        logging.debug('Unknown service registration passphrase! (%s != %s)' % (str(fields[1]),service_registration_passphrase))
        return False;
    elif fields[1] == "0":
        logging.debug("registration failed")
        return False
    else:
        logging.debug("else we go on:")        
        # start both heartbeat and listener
        service_connection_info['service_client_heartbeat_addr']=fields[2]
        service_connection_info['sender_id']=sender_id
        
        service_info = _get_service_info(application, service_id, sender_id, 0)
        if is_reregistration == 1 or not service_info is None:
            # we will need to re-connect
            # first, our service base in/out sockets
            #svc_conn = application.msg_conn
            #logging.debug("RECONNECT svc_conn (OUT:IN:SENDER) (%s:%s:%s)" % (svc_conn.out_addr, svc_conn.in_addr, sndr_id))
            #svc_conn.reconnect(sndr_id)

            if not service_info is None:
                service_conn = service_info['service_conn']
                #logging.debug("RECONNECT service_conn (OUT:IN:SENDER) (%s:%s:%s)" % (service_conn.out_addr, service_conn.in_addr, sndr_id))
                #service_conn.reconnect(sndr_id)
            else:
                service_connection_info['is_reregistration'] = 0
        else:
            if not service_heartbeat_addr is None:
                service_heartbeat_connection = ServiceHeartbeatConnection(application, **service_connection_info)
                application.msg_conn.set_service_heartbeat_connection(application, service_heartbeat_connection)

    return True

def start_service_client_registration_listener(application, 
    service_registration_passphrase, service_id, service_registration_addr, 
    service_client_heartbeat_addr = None, service_client_heartbeat_interval = DEFAULT_HEARTBEAT_INTERVAL):
    """public function to start our registration listeners for services to register with."""
    coro_spawn(_service_client_registration_listener, application, 'registration_listener',
        service_registration_passphrase, service_id, service_registration_addr, 
        service_client_heartbeat_addr, service_client_heartbeat_interval)
    # allow other events to process to make sure the listener is started
    coro_sleep(0)

def _service_client_registration_listener(application, 
    service_registration_passphrase, service_id, service_registration_addr, 
    service_client_heartbeat_addr = None, service_client_heartbeat_interval = DEFAULT_HEARTBEAT_INTERVAL):
    """Function runs in a coroutine, one registration listener for each brubeck instance
    When a registration message is received the service is registered and a listener is started.
    """
    logging.debug("_service_client_registration_listener: (service_registration_addr, service_registration_passphrase,service_client_heartbeat_interval): (%s, %s, %s)" % 
        (service_registration_addr, service_registration_passphrase,service_client_heartbeat_interval))
    # Just start our zmq socket here, no reason for abstraction
    zmq = load_zmq()
    ctx = load_zmq_ctx()
    service_registration_sock = ctx.socket(zmq.REP)
    print("Binding service registration listener REP socket %s" % (service_registration_addr))
    service_registration_sock.bind(service_registration_addr)
    print("BOUND service registration listener REP socket %s" % (service_registration_addr))

    while True:
        logging.debug("_service_client_registration_listener waiting");
        raw_registration_request = service_registration_sock.recv()
        logging.debug("_service_client_registration_listener recv(): %s" % raw_registration_request)
        # 21a6d348083f4b7d8d9680be387e01da 29:my_shared_registration_secret 8:run_slow 14:ipc://run/slow 23:ipc://run/slow_response 16:my_shared_secret 27:ipc://run/service_heartbeat
        # just send raw message to connection client
        fields = (sender_id, service_reg_passphrase, service_id, 
            service_addr, service_resp_addr, service_passphrase, heartbeat_addr
        ) = raw_registration_request.split(' ', 6)
        i=1
        for field in fields[1:]:
            fields[i] = t_parse(field)[0]
            i+=1
        # our minimal "security"    
        if fields[1] != service_registration_passphrase:
            # just log the breach
            logging.debug('Unknown service registration passphrase! (%s != %s)' % (str(fields[1]),service_registration_passphrase))
            msg = "%s %s %s" % (t(fields[1]), t(0), t(service_client_heartbeat_addr))
        else:
            service_conn = ServiceClientConnection(application, fields[3], fields[4], fields[5], sender_id)
            
            logging.debug("_service_client_registration_listener sender_id: %s" % sender_id)
            _register_service(application, 
                fields[5], #serv_passphrase
                fields[2], #service_id 
                fields[3], # service_addr
                fields[4], # service_resp_addr
                service_conn)
            
            service_connection_info = {
                    'service_registration_passphrase': fields[5],
                    'service_id': fields[2],
                    'service_registration_addr': service_registration_addr,
                    'service_passphrase': fields[0],    
                    'service_addr': fields[3],
                    'service_response_addr': fields[4],
                    'service_heartbeat_addr': fields[6],
                    'service_client_heartbeat_addr': service_client_heartbeat_addr,
                    'service_heartbeat_timeout': _DEFAULT_SERVICE_TIMEOUT,
                    'service_heartbeat_interval': service_client_heartbeat_interval,
                    'sender_id': sender_id,
                    'is_reregistration': 0,
            }
            if not service_client_heartbeat_addr is None:
                client_heartbeat_connection = ClientHeartbeatConnection(application, 
                    **service_connection_info)
                service_conn.set_client_heartbeat_connection(application, client_heartbeat_connection)
            msg = "%s %s %s" % (sender_id, t(fields[1]), t(service_client_heartbeat_addr))
        coro_sleep(0)
        service_registration_sock.send(to_bytes(msg))

def parse_service_response(msg, passphrase):
    """Function for constructing a Reponse instance out of a
    message read straight off a zmq socket from a ServiceConnection.
    """
    logging.debug("parse_service_response: %s" % msg)

    fields = (sender, conn_id, request_timestamp, start_timestamp, end_timestamp, 
    msg_passphrase, origin_sender_id, origin_conn_id, origin_out_addr, handle_response, 
    path, method, rest) = msg.split(' ', 12)
    # first field is not tnetstring, no need to do anything
    # last is group of tnetstrings, will handle after
    i=1
    for field in fields[1:-1]:
        fields[i] = t_parse(field)[0]
        i+=1   
    # our minimal "security"    
    if fields[5] != passphrase:
        raise Exception('Unknown service identity! (%s != %s)' % (str(fields[5]),str(passphrase)))
    # deal with the "body" or rest that is group of tnetstrings
    (status_code, rest) = t_parse(rest)
    (status_msg, rest) = t_parse(rest)
    (arguments, rest) = t_parse(rest)
    (headers, rest) = t_parse(rest)
    (body, rest) = t_parse(rest)
    logging.debug("arguments: %s" % arguments)
    logging.debug("headers: %s" % headers)
    logging.debug("body: %s" % body)
    arguments = json.loads(arguments) if len(arguments) > 0 else {}
    headers = json.loads(headers) if len(headers) > 0 else {}
    if body[0] == "{":
        body = json.loads(body) if len(body) > 0 else {}
    else:
        body = {
            "RETURN_DATA": body,
        }
    service_response = ServiceResponse(**{
        "sender": fields[0], 
        "conn_id": fields[1], 
        "request_timestamp": fields[2],
        "start_timestamp": fields[3],
        "end_timestamp": fields[4],
        "origin_sender_id": fields[6], 
        "origin_conn_id": fields[7], 
        "origin_out_addr": fields[8], 
        "handle_response": fields[9],
        "path": fields[10],
        "method": fields[11], 
        "status_code": int(status_code), 
        "status_msg": status_msg,
        "arguments": arguments, 
        "headers": headers, 
        "body": body, 
    })
    return service_response


class HeartbeatConnection(Connection):
    """a HeartbeatConnection is a pair of distintly separate sockets.
    PING socket is a heartbeat set at a regular interval.
    PONG socket is a listener for a heartbeat.
    The two PING and PONG in this class should not be PING/PONGing each other.
    The use case is for a service and serivce client o be PING/PONGing each other both ways.
    """

    def __init__(self, application, *args, **kwargs):
        
        zmq = load_zmq()
        ctx = load_zmq_ctx()
        
        if kwargs['sender_id'] is None:
            self.sender_id = application.msg_conn.sender_id
        else:
            self.sender_id = kwargs['sender_id']

        self.application = application
        self.service_registration_passphrase = kwargs['service_registration_passphrase']
        self.service_id = kwargs['service_id']
        self.service_registration_addr = kwargs['service_registration_addr']
        self.service_passphrase=kwargs['service_passphrase']
        self.service_addr=kwargs['service_addr']
        self.service_response_addr=kwargs['service_response_addr']
        self.service_heartbeat_addr=kwargs['service_heartbeat_addr']
        self.service_client_heartbeat_addr=kwargs['service_client_heartbeat_addr']
        self.service_heartbeat_timeout = kwargs['service_heartbeat_timeout']
        self.service_heartbeat_interval = kwargs['service_heartbeat_interval']
        self.sender_id = kwargs['sender_id']
        self.is_reregistration= kwargs.get('is_reregistration', 0)
        self.heartbeat= kwargs.get('heartbeat', False)
        #self.sender_id = uuid4().hex
                    
        self.in_sock = ctx.socket(zmq.SUB) # our listener
        self.out_sock = ctx.socket(zmq.PUB) # our heartbeat

        self.in_sock.setsockopt(zmq.SUBSCRIBE, self.sender_id)

    def get_service_connection_info(self):
        return {
            'service_registration_passphrase': self.service_registration_passphrase,
            'service_id': self.service_id,
            'service_registration_addr': self.service_registration_addr,
            'service_passphrase': self.service_passphrase,
            'service_addr': self.service_addr,
            'service_response_addr': self.service_response_addr,
            'service_heartbeat_addr': self.service_heartbeat_addr,
            'service_client_heartbeat_addr': self.service_client_heartbeat_addr,
            'service_heartbeat_timeout': self.service_heartbeat_timeout,
            'service_heartbeat_interval': self.service_heartbeat_interval,
            'sender_id': self.sender_id,
            'is_reregistration': (1 if self.is_reregistration == 1 else 0),
            'heartbeat': self.heartbeat,
        }

    def start(self, application):
        def start_me_up(application):
            self.beat_forever(application)
            self.listen_forever(application)

        coro_spawn(start_me_up, application)

    def recv(self):
        """Receives a raw mongrel2.handler.Request object that you from the
        zeromq socket and return whatever is found.
        """
        zmq_msg = self.in_sock.recv()
        return zmq_msg
        
    def close(self):
        """Tells mongrel2 to explicitly close the HTTP connection.
        """
        self.alive = False

    def listen_forever(self, application):
        """Defines a function that will be the heartbeat listener.
        """
        pass

    def beat_forever(self, application):
        """our heartbeat"""
        pass
    
    def send(self, sender_id, msg):
        """send our heartbeat itseld"""
        self.out_sock.send('%s %s' % (sender_id, to_bytes(msg)) );


class ServiceHeartbeatConnection(HeartbeatConnection):
    """a HeartbeatConnection is a pair of distintly separate sockets.
    PING socket is a heartbeat set at a regular interval.
    PONG socket is a listener for a heartbeat.
    The two PING and PONG in this class should not be PING/PONGing each other.
    The use case is for a service and serivce client o be PING/PONGing each other both ways.
    """

    def __init__(self, application, *args, **kwargs):            

        super(ServiceHeartbeatConnection, self).__init__(application, 
        *args, **kwargs)

        zmq = load_zmq()
        ctx = load_zmq_ctx()

        self.in_addr = self.service_client_heartbeat_addr
        self.out_addr = self.service_heartbeat_addr
        self.alive = True

        logging.debug("ServiceHeartbeatConnection (%s, %s)" % (self.in_addr ,self.out_addr ))
        
        self.in_sock.connect(self.in_addr)
        self.out_sock.setsockopt(zmq.IDENTITY, self.sender_id)
        self.out_sock.bind(self.out_addr)
        
        

    def listen_forever(self, application):
        """Defines a function that will be the heartbeat listener.
        """
        def fun_listen_forever(application):
            # forever is a long time, let's have a timeout
            zmq = load_zmq()
            ctx = load_zmq_ctx()

            poller = zmq.Poller()
            poller.register(self.in_sock, zmq.POLLIN)

            logging.debug("ServiceHeartbeatConnection listen_forever (%s, %s)." % (self.in_addr, self.sender_id))
            while self.alive:
                logging.debug("ServiceHeartbeatConnection listen_forever ... %s" % self.service_heartbeat_timeout)
                socks = dict(poller.poll(self.service_heartbeat_timeout * 1000))
                raw_heartbeat_response = None
                if socks and socks.get(self.in_sock) == zmq.POLLIN:
                    raw_heartbeat_response = self.in_sock.recv()
                    logging.debug("service PONG (%s,%s)" % (self.in_addr, self.sender_id))
                    coro_sleep(0)

                if raw_heartbeat_response is None:
                    logging.debug("service PONG TIMEOUT in service client heartbeat listener (%s,%s)" % (self.in_addr, self.sender_id))
                    # let us all die
                    self.alive=False
                    kwargs = self.get_service_connection_info()

                    # Just reregister all over again
                    kwargs['is_reregistration'] = 1

                    service_conn = ServiceClientConnection(application, 
                        kwargs['service_addr'], 
                        kwargs['service_response_addr'], 
                        kwargs['service_passphrase'])

                    logging.debug("ServiceHeartbeatConnection listen_forever self.sender_id: %s" % self.sender_id)
                    _register_service(application, 
                       kwargs['service_passphrase'], #serv_passphrase
                       kwargs['service_id'], #service_id 
                       kwargs['service_addr'], # service_addr
                       kwargs['service_response_addr'], # service_resp_addr
                       service_conn)

                    service_heartbeat_connection = ServiceHeartbeatConnection(application, 
                        **kwargs)
                    application.msg_conn.set_service_heartbeat_connection(application, service_heartbeat_connection)

            self.in_sock.close()
        coro_spawn(fun_listen_forever, application, 'client_heartbeat_listener')

    def beat_forever(self, application):
        """our heartbeat"""
        def the_beat_is_on(application):
            logging.debug("SERVICE beat is on")
            while self.alive:
                msg = "%s %s %s %s" % (self.sender_id, t(self.service_passphrase), t(self.service_id), t(self.service_heartbeat_addr))
                self.send(self.sender_id, msg)
                logging.debug("service PING (%s:%s)" % (self.out_addr, self.sender_id))
                coro_sleep(self.service_heartbeat_interval)

            if not self.alive:
                logging.debug("service PING DOWN self.alive check failed(%s:%s)" % (self.out_addr, self.sender_id))
            self.out_sock.close()
            
        logging.debug("SERVICE beat_forever spawning the_beat_is_on")
        coro_spawn(the_beat_is_on, application, 'service_heartbeat')


class ClientHeartbeatConnection(HeartbeatConnection):
    """a HeartbeatConnection is a pair of distintly separate sockets.
    PING socket is a heartbeat set at a regular interval.
    PONG socket is a listener for a heartbeat.
    The two PING and PONG in this class should not be PING/PONGing each other.
    The use case is for a service and serivce client o be PING/PONGing each other both ways.
    """

    def __init__(self, application, *args, **kwargs):            

        super(ClientHeartbeatConnection, self).__init__(application, 
        *args, **kwargs)

        zmq = load_zmq()
        ctx = load_zmq_ctx()

        self.in_addr = self.service_heartbeat_addr
        self.out_addr = self.service_client_heartbeat_addr
        self.alive = True
      

        logging.debug("ClientHeartbeatConnection (%s, %s)" % (self.in_addr ,self.out_addr ))
        
        self.in_sock.connect(self.in_addr)
        self.out_sock.setsockopt(zmq.IDENTITY, self.sender_id)
        self.out_sock.bind(self.out_addr)


    def listen_forever(self, application):
        """Defines a function that will be the heartbeat listener.
        """
        def listen_forever(application):
            # forever is a long time, let's have a timeout
            zmq = load_zmq()
            ctx = load_zmq_ctx()
        
            poller = zmq.Poller()
            poller.register(self.in_sock, zmq.POLLIN)
        
            logging.debug("ClientHeartbeatConnection listen_forever (%s, %s)." % (self.in_addr, self.sender_id))
            while self.alive:
                logging.debug("ClientHeartbeatConnection listen_forever ... %s" % self.service_heartbeat_timeout)
                socks = dict(poller.poll(self.service_heartbeat_timeout * 1000))
                raw_heartbeat_response = None
                if socks and socks.get(self.in_sock) == zmq.POLLIN:
                    raw_heartbeat_response = self.in_sock.recv()
                    logging.debug("client PONG (%s, %s)" % (self.in_addr, self.sender_id))
                    coro_sleep(0)
                if raw_heartbeat_response is None:
                    logging.debug("client PONG TIMEOUT in service heartbeat listener (%s,%s)" % (self.in_addr, self.sender_id))
                    self.alive=False
                    #unreregister our service from the application
                    logging.debug("_unregister_service (%s,%s,%s)" % (self.application.msg_conn.sender_id, self.service_id, self.sender_id)) 
                    _unregister_service(self.application, self.service_id, self.sender_id)
                    
                    # reregister our service
                    #_service_registration(application, 
                    #    self.service_passphrase, self.service_id, self.service_addr,
                    #    self.service_passphrase, self.service_addr, self.service_response_addr,  
                    #    self.service_heartbeat_addr, self.service_client_heartbeat_addr, 
                    #    self.service_heartbeat_timeout, self.sender_id,
                    #    1)
                             
        coro_spawn(listen_forever, application, 'service_heartbeat_listener')

    def beat_forever(self, application):
        """our heartbeat"""
        def the_beat_is_on(application):
            logging.debug("CLIENT beat is on")
            while self.alive:
                msg = "%s %s %s %s" % (self.sender_id, t(self.service_passphrase), t(self.service_id), t(self.service_heartbeat_addr))
                self.send(self.sender_id, msg)
                logging.debug("client PING (%s:%s)" % (self.out_addr, self.sender_id))
                coro_sleep(self.service_heartbeat_interval)
            if not self.alive:
                logging.debug("client PING DOWN self.alive check failed(%s:%s)" % (self.out_addr, self.sender_id))

        logging.debug("CLIENT beat_forever spawning the_beat_is_on")
        coro_spawn(the_beat_is_on, application, 'client_heartbeat')
